import networkx as nx
from typing import *

def dynamicScore(U : set, V : set) -> float:
    """
    Returns the dynamic score of the two sets of nodes U and V.
    """
    return 0 if len(U) == 0 or len(V) == 0 else 1 - len(U.intersection(V)) / len(U.union(V))

def verticesDynamicScore(G1 : nx.Graph, G2 : nx.Graph) -> float:
    """
    Returns the Vertices-DynamicScore of the two graphs G1 and G2.
    """
    return dynamicScore(set(G1.nodes), set(G2.nodes))

def edgesDynamicScore(G1 : nx.Graph, G2 : nx.Graph) -> float:
    """
    Returns the Edges-DynamicScore of the two graphs G1 and G2.
    """
    return dynamicScore(set(G1.edges), set(G2.edges))

def listVerticesDynamicScore(Gs : List[nx.Graph]) -> List[float]:
    """
    Returns the list Vertices-DynamicScore of the two graphs G1 and G2.
    """
    return list(dynamicScore(set(Gs[0].nodes), set(Gs[1].nodes)) for Gs in zip(Gs, Gs[1:]))

def listEdgesDynamicScore(Gs : List[nx.Graph]) -> List[float]:
    """
    Returns the list Edges-DynamicScore of the two graphs G1 and G2.
    """
    return list(dynamicScore(set(Gs[0].edges), set(Gs[1].edges)) for Gs in zip(Gs, Gs[1:]))