import networkx as nx
from typing import *

def cumulativeEdgeGraph(Gs : List[nx.Graph]):
    res = []
    G = nx.Graph()
    for i, Gi in enumerate(Gs):
        # Add the vertices
        G.add_nodes_from(Gi.nodes()) # Add from wont add if the node already exists
        # Add the edges if they don't exist and add the weight if they do
        for u, v in Gi.edges():
            if G.has_edge(u, v):
                G[u][v]['weight'] += 1
            else:
                G.add_edge(u, v)
                G[u][v]['weight'] = 1
        res.append(G.copy())
    return res

def growingGraph(Gs : List[nx.Graph]):
    res = []
    G = nx.Graph()
    for i, Gi in enumerate(Gs):
        G.add_nodes_from(Gi.nodes())
        G.add_edges_from(Gi.edges())
        res.append(G.copy())
    return res

def buildDynamicGraphSlidingWindowAtt(GCs : List[nx.Graph], window : int, t):
    res = GCs[t].copy()
    if t - window >= 0:
        GtoRemove = GCs[t - window]
        for u, v in GtoRemove.edges():
            if res.has_edge(u, v):
                res[u][v]['weight'] -= GtoRemove[u][v]['weight']
                if res[u][v]['weight'] <= 0:
                    res.remove_edge(u, v)
        # Remvoe the nodes that have no edges
        res.remove_nodes_from(list(nx.isolates(res)))
    return res

def optimizedDynamicGraphSlidingWindow(Gs : List[nx.Graph], window : int):
    res = []
    T = len(Gs)
    G = nx.Graph()
    Gcummulative = cumulativeEdgeGraph(Gs)
    # Build dynamic graph using the cumulative edge graph
    for t in range(window, T):
        print(f"\r{t-window}/{T-window}", end="")
        res.append(buildDynamicGraphSlidingWindowAtt(Gcummulative, window, t))
    return res

def buildFirstToLastVertices(Gs : List[nx.Graph]):
    verticesFirstAndLastApparition = {}
    for i, G in enumerate(Gs):
        for v in G.nodes():
            if v not in verticesFirstAndLastApparition:
                verticesFirstAndLastApparition[v] = (i, i)
            else:
                verticesFirstAndLastApparition[v] = (verticesFirstAndLastApparition[v][0], i)
    return verticesFirstAndLastApparition

def buildFirstToLastEdges(Gs : List[nx.Graph]):
    edgesFirstAndLastApparition = {}
    for i, G in enumerate(Gs):
        for u, v in G.edges():
            if (u, v) not in edgesFirstAndLastApparition:
                edgesFirstAndLastApparition[(u, v)] = (i, i)
            else:
                edgesFirstAndLastApparition[(u, v)] = (edgesFirstAndLastApparition[(u, v)][0], i)
    return edgesFirstAndLastApparition

def firstToLast(Gs : List[nx.Graph]):
    verticesFirstAndLastApparition = buildFirstToLastVertices(Gs)
    allVertices = set(verticesFirstAndLastApparition.keys())
    res = []
    # Each graph will have vertices if they appear between the first and last graph
    for i, G in enumerate(Gs):
        H = nx.Graph()
        H.add_edges_from(G.edges())
        H.add_nodes_from(G.nodes())
        for vertex in allVertices:
            if verticesFirstAndLastApparition[vertex][0] <= i <= verticesFirstAndLastApparition[vertex][1]:
                H.add_node(vertex)
        res.append(H.copy())
    return res

def firstToLastVerticesAndEdges(Gs : List[nx.Graph]):
    verticesFirstAndLastApparition = buildFirstToLastVertices(Gs)
    allVertices = set(verticesFirstAndLastApparition.keys())
    edgesFirstAndLastApparition = buildFirstToLastEdges(Gs)
    allEdges = set(edgesFirstAndLastApparition.keys())
    res = []
    # Each graph will have vertices and edges if they appear between the first and last graph
    for i, G in enumerate(Gs):
        H = nx.Graph()
        H.add_edges_from(G.edges())
        H.add_nodes_from(G.nodes())
        for vertex in allVertices:
            if verticesFirstAndLastApparition[vertex][0] <= i <= verticesFirstAndLastApparition[vertex][1]:
                H.add_node(vertex)
        for edge in allEdges:
            if edgesFirstAndLastApparition[edge][0] <= i <= edgesFirstAndLastApparition[edge][1]:
                H.add_edge(*edge)
        res.append(H.copy())
    return res

def quickUnion(G : nx.Graph, H : nx.Graph) -> None:
    G.add_edges_from(H.edges)

def timeInterval(Gs : List[nx.Graph], window : int):
    res = []
    T = len(Gs)
    for t in range(window, T, window):
        print(f"\r{t-window}/{T-window}", end="")
        G = Gs[t].copy()
        for i in range(t, min(T, t+window)):
            quickUnion(G, Gs[i])
        res.append(G)
    return res

def slidingWindowMetricsOnly(Gs : List[nx.Graph], window : int, metrics1d : List[Callable[[nx.Graph], float]], metrics2d : List[Callable[[nx.Graph, nx.Graph], float]]):
    Gtmp = Gs[0].copy()
    for t in range(1, window):
        quickUnion(Gtmp, Gs[t])
    res = []
    T = len(Gs)
    for t in range(window, T):
        print(f"\r{t-window}/{T-window}", end="")
        G = Gs[t].copy()
        for i in range(t-window, t):
            quickUnion(G, Gs[i])
        res.append([metric(G) for metric in metrics1d] + [metric(G, Gtmp) for metric in metrics2d])
        Gtmp = G
    return res